// Linux Device Driver Template/Skeleton
// Kernel Module

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include "x86_superio.h"

//#define MAJOR 104
//#define NAME "dio"
#define MOXA_DIO_MINOR 104
//#define BASEPORT 0x800
#define BASEPORT 0x300

//
// DIO file operaiton function call
//
#define MAX_DIO                 4

#define DIO_INPUT               1
#define DIO_OUTPUT              0
#define DIO_HIGH                1
#define DIO_LOW                 0
#define IOCTL_DIO_GET_MODE      1
#define IOCTL_DIO_SET_MODE      2
#define IOCTL_DIO_GET_DATA      3
#define IOCTL_DIO_SET_DATA      4
#define IOCTL_SET_DOUT          15
#define IOCTL_GET_DOUT          16
#define IOCTL_GET_DIN           17

#define MOXA                    0x400
#define MOXA_SET_OP_MODE      (MOXA + 66)
#define MOXA_GET_OP_MODE      (MOXA + 67)
//#define MOXA_SET_OP_MODE      (MOXA + 100)
//#define MOXA_GET_OP_MODE      (MOXA + 101)

#define RS232_MODE              0
#define RS485_2WIRE_MODE        1
#define RS422_MODE              2
#define RS485_4WIRE_MODE        3


struct dio_set_struct {
        int     io_number;
        int     mode_data;      // 1 for input, 0 for output, 1 for high, 0 for low
};

static unsigned char do_state_keep=0xff;
static unsigned int counter = 0;
static char string [128];

// open function - called when the "file" /dev/v2100_io is opened in userspace
static int io_open (struct inode *inode, struct file *file) {
	//printk("io_open\n");
	return 0;
}

// close function - called when the "file" /dev/v2100_io is closed in userspace  
static int io_release (struct inode *inode, struct file *file) {
	//printk("io_release\n");
	return 0;
}

// read function called when from /dev/v2100_io is read
static ssize_t io_read (struct file *file, char *buf,
		size_t count, loff_t *ppos) {
	int len, err;
	
	if( counter <= 0 ) 
		return 0;
	err = copy_to_user(buf,string,counter);
	if (err != 0)
		return -EFAULT;
	len  = counter;
	counter = 0;
	return len;
	return 0;
}

// write function called when to /dev/v2100_io is written
static ssize_t io_write (struct file *file, const char *buf,
		size_t count, loff_t *ppos) {
	int err;
	err = copy_from_user(string,buf,count);

	if(count < 3)
		return -EINVAL;
	if (err != 0)
		return -EFAULT;

	outb((unsigned char)string[2], (((unsigned short)string[0])<<8)|((unsigned short)string[1]));

	counter += count;
	return count;
}

unsigned char keep_opmode=0x00;
// ioctl - I/O control
//static int io_ioctl(struct inode *inode, struct file *file,unsigned int cmd, unsigned long arg) {
static long io_ioctl( struct file *file,unsigned int cmd, unsigned long arg) {
	struct dio_set_struct   set;
	unsigned char di_state;
	unsigned char port,opmode,val;

	//printk("get op mode\n");
	
	switch ( cmd ) {
	case IOCTL_SET_DOUT :
		if ( copy_from_user(&set, (struct dio_set_struct *)arg, sizeof(struct dio_set_struct)) )
			return -EFAULT;
                if ( set.io_number < 0 || set.io_number >= MAX_DIO )
                        return -EINVAL;
                if ( set.mode_data == DIO_HIGH )
                        do_state_keep |= (1<<set.io_number);
                else if ( set.mode_data == DIO_LOW )
                        do_state_keep &= ~(1<<set.io_number);
                else
                        return -EINVAL;
		outb(do_state_keep,BASEPORT+5);
		//printk("%s,%d :BASEPORT+5:%x\n", __FUNCTION__, __LINE__,(unsigned long)inb(BASEPORT+5));	
		break;
        case IOCTL_GET_DOUT :
        case IOCTL_GET_DIN :
                if ( copy_from_user(&set, (struct dio_set_struct *)arg, sizeof(struct dio_set_struct)) )
                        return -EFAULT;
		if ( set.io_number == -1 ) {	// to get all port
			if ( cmd == IOCTL_GET_DOUT )
				set.mode_data = do_state_keep & 0xf;
			else
				set.mode_data = (inb(BASEPORT+5)>>4) & 0xf;
			goto ioctl_get_label;
		}
		if ( set.io_number < 0 || set.io_number >= MAX_DIO )
			return -EINVAL;
		if ( cmd == IOCTL_GET_DOUT ) {
			if ( do_state_keep & (1<<set.io_number) )
				set.mode_data = 1;
			else
				set.mode_data = 0;
		} else {
			di_state = inb(BASEPORT+5)>>4;
			if ( di_state & (1<<set.io_number) )
				set.mode_data = 1;
			else
				set.mode_data = 0;
		}
ioctl_get_label:
		if ( copy_to_user((struct dio_set_struct *)arg, &set, sizeof(struct dio_set_struct)) )
			return -EFAULT;
		//printk("%s,%d :BASEPORT+5:%x\n", __FUNCTION__, __LINE__,(unsigned long)inb(BASEPORT+5));
		break;
	case MOXA_SET_OP_MODE:
		copy_from_user(&opmode,(unsigned char *)arg,sizeof(unsigned char));
		port = opmode >> 4 ;
		opmode = opmode & 0xf;

		if ( opmode != RS232_MODE && opmode != RS485_2WIRE_MODE && opmode != RS422_MODE && opmode != RS485_4WIRE_MODE && port > 1)
		return -EFAULT;

		//printk("port:%x,opmode:%x\n",port,opmode);
		val=inb(BASEPORT+4)&(~(((unsigned char)0xe)<<(4*port)));
		//printk("val:%x\n",val);

		switch(opmode){
		case RS232_MODE:
			val|=(((unsigned char)0x8)<<(4*port));
			break;			
		case RS485_2WIRE_MODE:
			val|=(((unsigned char)0x2)<<(4*port));
			break;
		case RS422_MODE:
		case RS485_4WIRE_MODE:
			val|=(((unsigned char)0x4)<<(4*port));
			break;
		}

		//printk("val:%x\n",val);
		outb(val,BASEPORT+4);

		keep_opmode &= ~(((unsigned char)0xf)<<(port*4));
		keep_opmode |= opmode<<(port*4);

		superio_enter_config();
		superio_set_logic_device((u8)(port+1));
		if(opmode == RS232_MODE){
			val=superio_get_reg(0xf0)& 0x7f;
		}else{
			val=superio_get_reg(0xf0)| 0x80;
		}
		superio_set_reg(val , 0xf0);

		break;

	case MOXA_GET_OP_MODE:
		//printk("get op mode\n");
		copy_from_user(&port,(unsigned char *)arg,sizeof(unsigned char));
		if(port>1)return -EINVAL;
		opmode=(keep_opmode>>(port*4))& 0xf;
		copy_to_user((unsigned char*)arg, &opmode, sizeof(unsigned char));

		break;
	default:
		//printk("ioctl:error\n");
		return -EINVAL;
	}
	return 0;
}

#if 0
// define which file operations are supported
struct file_operations v2100_io_fops = {
	.owner	=	THIS_MODULE,
	.llseek	=	NULL,
	.read		=	NULL,/*io_read,*/
	.write	=	NULL,/*io_write,*/
	.readdir	=	NULL,
	.poll		=	NULL,
	.ioctl	=	io_ioctl,
	.mmap		=	NULL,
	.open		=	io_open,
	.flush	=	NULL,
	.release	=	io_release,
	.fsync	=	NULL,
	.fasync	=	NULL,
	.lock		=	NULL,
};
#endif

// define which file operations are supported
struct file_operations io_fops = {
	//.owner	=	THIS_MODULE,
	//.llseek	=	NULL,
	//.read		=	NULL,/*io_read,*/
	//.write	=	NULL,/*io_write,*/
	//.readdir	=	NULL,
	//.poll		=	NULL,
	//.ioctl	=   io_ioctl,
	//.compat_ioctl	=   io_ioctl,
	.unlocked_ioctl	=   io_ioctl,
	//.mmap		=	NULL,
	.open		=   io_open,
	//.flush	=	NULL,
	.release	=   io_release,
	//.fsync	=	NULL,
	//.fasync	=	NULL,
	//.lock		=	NULL,
};

// register as misc driver
static struct miscdevice dio_miscdev = {
    .minor = MOXA_DIO_MINOR,
    .name = "dio",
    .fops = &io_fops,
};


// initialize module (and interrupt)
static int __init io_init_module (void) {
	
	int i;
	//int ret;
	unsigned char val;
	printk("initializing MOXA DIO module\n");
	
	//i = register_chrdev (MAJOR, NAME, &v2100_io_fops);
	//if (i != 0) return - EIO;
	if ( misc_register(&dio_miscdev)!=0 ) {
        	printk("Moxa dio driver: Register misc fail !\n");
        	return -ENOMEM;
	}
/*
	outb(do_state_keep, BASEPORT+5);
	outb(0x00, BASEPORT);
	
	//set default serial mode to RS232
	//outb(0x88, BASEPORT+4);
	outb(0x99, BASEPORT+4);

	superio_enter_config();
	superio_set_logic_device(1);
	val=superio_get_reg(0xf0)&0x7f;
	superio_set_reg(val,0xf0);

	superio_enter_config();
	superio_set_logic_device(2);
	val=superio_get_reg(0xf0)&0x7f;
	superio_set_reg(val,0xf0);
*/	
	return 0;
}

// close and cleanup module
static void __exit io_cleanup_module (void) {
	printk("cleaning up module\n");
	//unregister_chrdev (MAJOR, NAME);
	misc_deregister(&dio_miscdev);
}

module_init(io_init_module);
module_exit(io_cleanup_module);
MODULE_AUTHOR("johnson_liu@moxa.com");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("MOXA DIO module");

