/*
 * superio.c -- a module for superio control 
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h> /* printk() */
#include <linux/fs.h>     /* everything... */
#include <linux/types.h>  /* size_t */
#include <linux/spinlock.h>
#include <linux/ioport.h>

#include <asm/io.h>
#include <linux/ioport.h>

#define	SUPERIO_CONFIG_PORT		0x2e

static spinlock_t superio_lock = SPIN_LOCK_UNLOCKED;


/* jackie use "asm out 0xeb, al" as delay, it's the same way */
static inline void delay() {
	outb( 0x80,0xeb); 
}

static void set_reg(u8 val,u8 index) {
	outb(index, SUPERIO_CONFIG_PORT);
	delay();
	outb(val, (SUPERIO_CONFIG_PORT+1));
	delay();
}

static void enter_config(u8 logic_dev) {
/*
#if defined (DA68X) || defined (DA710)
	outb (0x87, SUPERIO_CONFIG_PORT);
	delay();
	outb (0x87, SUPERIO_CONFIG_PORT);
	delay();
	set_reg(logic_dev, 0x07);
	delay();
#elif defined (V21XX)
*/
	outb (0x87, SUPERIO_CONFIG_PORT);
	outb (0x01, SUPERIO_CONFIG_PORT);
	outb (0x55, SUPERIO_CONFIG_PORT);
	outb (0x55, SUPERIO_CONFIG_PORT);
	delay();
	set_reg(logic_dev, 0x07);
	delay();
//#endif
}

static void exit_config(void) {
/*
#if defined (DA68X) || defined (DA710)
	outb (0xaa, SUPERIO_CONFIG_PORT);
	delay();
#elif defined (V21XX)
*/
	outb(0x02, SUPERIO_CONFIG_PORT);
	delay();
	outb(0x02, SUPERIO_CONFIG_PORT+1);
//#endif
}

/* 
 * wade: use spin_lock_irqsave to prevent the interrupt 
 * from hw interrupt /system call/ softirq 
 */
unsigned char superio_read(u8 logic_dev, u8 index) 
{
	unsigned char val;
	unsigned long flags;

	spin_lock_irqsave(&superio_lock, flags);
	enter_config(logic_dev);

	/* get reg */
	outb (index, SUPERIO_CONFIG_PORT);
	delay();
	val = inb(SUPERIO_CONFIG_PORT + 1);
	delay();

	exit_config();
	spin_unlock_irqrestore(&superio_lock, flags);
	return val;
}
EXPORT_SYMBOL(superio_read);

void superio_write(u8 logic_dev, u8 index, u8 val) 
{
	unsigned long flags;

	spin_lock_irqsave(&superio_lock, flags);
	enter_config(logic_dev);

	/* set reg */
	outb(index, SUPERIO_CONFIG_PORT);
	delay();
	outb(val, SUPERIO_CONFIG_PORT + 1);
	delay();

	exit_config();
	spin_unlock_irqrestore(&superio_lock, flags);
}
EXPORT_SYMBOL(superio_write);
EXPORT_SYMBOL(superio_read);

int superio_init(void)
{
	printk("Load Moxa superio driver\n");
	if (!request_region(SUPERIO_CONFIG_PORT, 2, "moxa_superio")) {
		printk("Can't request region !");
		return -1;
	}
	return 0;
}

void superio_cleanup(void)
{
	printk("Unload Moxa superio driver\n");
	release_region(SUPERIO_CONFIG_PORT, 2);
}

module_init(superio_init);
module_exit(superio_cleanup);
MODULE_LICENSE("GPL");

